"""Supplicant-Facing Sockets"""

import struct
from abc import ABC, abstractmethod
from fcntl import ioctl
from eventlet.green import socket

from chewie.mac_address import MacAddress
from chewie.utils import get_logger


class PromiscuousSocket(ABC):
    """Abstract Raw Socket in Promiscuous Mode"""
    SIOCGIFINDEX = 0x8933
    PACKET_MR_PROMISC = 1
    SOL_PACKET = 263
    PACKET_ADD_MEMBERSHIP = 1
    EAP_ADDRESS = MacAddress.from_string("01:80:c2:00:00:03")

    @abstractmethod
    def send(self, data):  # pylint: disable=missing-docstring
        pass

    @abstractmethod
    def receive(self):  # pylint: disable=missing-docstring
        pass

    @abstractmethod
    def setup(self):  # pylint: disable=missing-docstring
        pass

    def __init__(self, interface_name, log_prefix):
        self.socket = None
        self.interface_index = None
        self.interface_name = interface_name
        self.logger = get_logger(log_prefix)

    def _setup(self, socket_filter):
        """Set up the socket"""
        self.logger.info("Setting up socket on interface: %s", self.interface_name)
        try:
            self.open(socket_filter)
            self.get_interface_index()
            self.set_interface_promiscuous()
        except socket.error as err:
            self.logger.error("Unable to setup socket: %s", str(err))
            raise err

    def open(self, socket_filter):
        """Setup EAP socket"""
        self.socket = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket_filter)
        self.socket.bind((self.interface_name, 0))

    def get_interface_index(self):
        """Get the interface index of the EAP Socket"""
        # http://man7.org/linux/man-pages/man7/netdevice.7.html
        request = struct.pack('16sI', self.interface_name.encode("utf-8"), 0)
        response = ioctl(self.socket, self.SIOCGIFINDEX, request)
        _ifname, self.interface_index = struct.unpack('16sI', response)

    def set_interface_promiscuous(self):
        """Sets the EAP interface to be able to receive EAP messages"""
        request = struct.pack("IHH8s", self.interface_index, self.PACKET_MR_PROMISC,
                              len(self.EAP_ADDRESS.address), self.EAP_ADDRESS.address)
        self.socket.setsockopt(self.SOL_PACKET, self.PACKET_ADD_MEMBERSHIP, request)


class EapSocket(PromiscuousSocket):
    """Handle the EAP socket"""

    def setup(self):
        """Set up the socket"""
        self._setup(socket.htons(0x888e))

    def send(self, data):
        """send on eap socket.
            data (bytes): data to send"""
        self.socket.send(data)

    def receive(self):
        """receive from eap socket"""
        return self.socket.recv(4096)


class MabSocket(PromiscuousSocket):
    """Handle the Mab socket for DHCP Requests"""
    IP_ETHERTYPE = 0x0800
    DHCP_UDP_SRC = 68
    DHCP_UDP_DST = 67
    UDP_IPTYPE = b'\x11'

    def setup(self):
        """Set up the socket"""
        self._setup(socket.htons(self.IP_ETHERTYPE))

    def send(self, data):
        """Not Implemented -- This socket is purely for Listening"""
        raise NotImplementedError('Attempted to send data down the activity socket')

    def receive(self):
        """Receive activity from supplicant-facing socket"""
        # Skip all packets that are not DHCP requests
        while True:
            ret_val = self.socket.recv(4096)

            if ret_val[23:24] == self.UDP_IPTYPE:
                src_port = struct.unpack('>H', ret_val[34:36])[0]
                dst_port = struct.unpack('>H', ret_val[36:38])[0]

                if src_port == self.DHCP_UDP_SRC and dst_port == self.DHCP_UDP_DST:
                    return ret_val



"""Test Chewie's EAP Functionality"""

import time
from base_test import BaseTest


class EapTest(BaseTest):
    """Test Chewie's EAP Functionality"""
    test_name = "EapTest"

    def setUp(self):
        """Start Radius and Chewie Servers"""
        super().setUp()
        self.start_radius()
        self.start_chewie()

    def test_peap(self):
        """Attempt to connect to Chewie using PEAP on WPA_Supplicant"""
        self.start_wpa_supplicant('peap')
        time.sleep(5)
        self.check_output()

    def test_md5(self):
        """Attempt to connect to Chewie using MD5 on WPA_Supplicant"""
        self.start_wpa_supplicant('md5')
        time.sleep(5)
        self.check_output()

    def test_tls(self):
        """Attempt to connect to Chewie using TLS on WPA_Supplicant"""
        self.start_wpa_supplicant('tls')
        time.sleep(5)
        self.check_output()

    def test_ttls(self):
        """Attempt to connect to Chewie using TTLS on WPA_Supplicant"""
        self.start_wpa_supplicant('ttls')
        time.sleep(5)
        self.check_output()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../'))


# -- Project information -----------------------------------------------------

project = 'Chewie'
copyright = '2019, Chewie Developers'
author = 'Chewie Developers'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.graphviz'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'README.rst', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Magic to run sphinx-apidoc automatically -----------------------------

# See https://github.com/rtfd/readthedocs.org/issues/1139
# See https://github.com/faucetsdn/faucet/blob/master/docs/conf.py
# on which this is based.

def run_apidoc(_):
    """Call sphinx-apidoc on chewie module"""
    from sphinx.ext.apidoc import main as apidoc_main
    apidoc_main(['-e', '-o', 'source/apidoc', '../chewie'])


def build_state_machine_diagrams(_):
    from chewie.state_machines.eap_state_machine import FullEAPStateMachine
    from chewie.state_machines.mab_state_machine import MacAuthenticationBypassStateMachine \
        as MABStateMachine

    FullEAPStateMachine.build_state_graph('eap_state_machine.png')
    MABStateMachine.build_state_graph('mab_state_machine.png')


def setup(app):
    """ Add hooks into Sphinx to change behaviour and autogen documentation """

    # Add custom css
    app.add_css_file("css/responsive-tables.css")
    # Override Sphinx setup to trigger sphinx-apidoc.
    app.connect('builder-inited', run_apidoc)
    # Build State Machine Graphs
    app.connect('builder-inited', build_state_machine_diagrams)



"""Mock TimerScheduler
"""


class FakeTimerJob:
    """Behaves like TimerJob"""

    def __init__(self, function, args, timeout):
        self.function = function
        self.args = args
        self.timeout = timeout
        self.is_cancelled = False

    def cancel(self):
        """Clones TimerJob.cancel()"""
        self.is_cancelled = True

    def cancelled(self):
        """Clones TimerJob.cancelled()"""
        return self.is_cancelled

    def run(self):
        """Runs job"""
        if not self.is_cancelled:
            self.function(*self.args)


class FakeTimerScheduler:
    """Behaves like TimerScheduler"""

    def __init__(self):
        self.jobs = []

    def call_later(self, timeout, func, *args):
        """Clones TimerScheduler.call_later()"""
        if not args:
            args = []

        job = FakeTimerJob(func, args, timeout)

        self.jobs.append(job)

        return job

    def run_jobs(self, num_jobs=None):
        """Runs jobs in order of timeout"""
        if not num_jobs:
            while self.jobs:
                self.jobs.sort(key=lambda x: x.timeout)
                job = self.jobs.pop(0)
                job.run()
                print("ddddddddd")
        else:
            for _ in range(num_jobs):
                if not self.jobs:
                    break
                self.jobs.sort(key=lambda x: x.timeout)
                job = self.jobs.pop(0)
                job.run()
                print("eeeeeeeeez")

    def run(self):
        """Clones TimerScheduler.run()"""
        pass



"""This Module provides a Bare-bones Mac Authentication Bypass State Machine provide 802.1x
MAB Support in Chewie"""

from transitions import State, Machine

from chewie.event import EventMessageReceived, EventRadiusMessageReceived
from chewie.radius import RadiusAccessAccept, RadiusAccessReject
from chewie.utils import get_logger, log_method, RadiusQueueMessage
from chewie.state_machines.abstract_state_machine import AbstractStateMachine


class MacAuthenticationBypassStateMachine(AbstractStateMachine):
    """This Class provides a Bare-bones Mac Authentication Bypass State Machine provide 802.1x
    MAB Support in Chewie"""

    # pylint: disable=too-many-instance-attributes
    # REASON: Due to class being State Machine - Requires a lot of instance attributes

    DEFAULT_SESSION_TIMEOUT = 3600  # Number of Seconds

    DISABLED = "DISABLED"
    ETH_RECEIVED = "ETH_RECEIVED"
    AAA_REQUEST = "AAA_REQUEST"
    AAA_IDLE = "AAA_IDLE"
    AAA_RECEIVED = "AAA_RECEIVED"
    AAA_SUCCESS = "AAA_SUCCESS"
    AAA_FAILURE = "AAA_FAILURE"

    INITIAL_STATE = DISABLED
    PROGRESS_STATES = [
        State(DISABLED, 'mab_disabled_state'),
        State(ETH_RECEIVED, 'eth_received_state'),
        State(AAA_REQUEST, 'aaa_request_state'),
        State(AAA_IDLE, 'aaa_idle_state'),
        State(AAA_RECEIVED, 'aaa_received_state'),
    ]

    SUCCESS_STATES = [State(AAA_SUCCESS, 'aaa_success_state'), ]
    FAILURE_STATES = [State(AAA_FAILURE, 'aaa_failure_state'), ]
    COMPLETION_STATES = FAILURE_STATES + SUCCESS_STATES

    STATES = COMPLETION_STATES + PROGRESS_STATES

    ERROR_TRANSTIONS = [
        {'trigger': 'process', 'source': '*', 'dest': DISABLED,
         'unless': ['_is_port_enabled']},
        {'trigger': 'process', 'source': '*', 'dest': DISABLED,
         'conditions': ['_is_mab_restart']},
    ]

    CORE_TRANSITIONS = [
        {'trigger': 'process', 'source': DISABLED, 'dest': ETH_RECEIVED,
         'conditions': ['_is_eth_received']},

        {'trigger': 'process', 'source': ETH_RECEIVED, 'dest': AAA_REQUEST},
        {'trigger': 'process', 'source': AAA_REQUEST, 'dest': AAA_IDLE},

        {'trigger': 'process', 'source': AAA_IDLE, 'dest': AAA_RECEIVED,
         'conditions': ['_is_aaa_received']},

        # Completion States
        {'trigger': 'process', 'source': AAA_RECEIVED, 'dest': AAA_FAILURE,
         'conditions': ['_is_aaa_fail']},
        {'trigger': 'process', 'source': AAA_RECEIVED, 'dest': AAA_SUCCESS,
         'conditions': ['_is_aaa_success']},

        # On Failure - Restart Authentication
        {'trigger': 'process', 'source': AAA_FAILURE, 'dest': ETH_RECEIVED,
         'conditions': ['_is_eth_received']},
    ]
    TRANSITIONS = CORE_TRANSITIONS + ERROR_TRANSTIONS

    state = None

    # State Variables
    port_enabled = False
    mab_restart = False
    aaa_received = False
    aaa_fail = False
    aaa_success = False

    aaa_response_data = None
    aaa_request_data = None
    aaa_response_attributes = None

    eth_received = True
    eth_message_data = None

    radius_state_attribute = None
    # NOTE: This is not dynamic at this stage. Session timeout Attributes from radius are ignored
    session_timeout = DEFAULT_SESSION_TIMEOUT
    port_id_mac = None

    #
    # State Transition Helpers
    #
    def _is_port_enabled(self):  # pylint: disable=missing-docstring
        return self.port_enabled

    def _is_eth_received(self):  # pylint: disable=missing-docstring
        return self.eth_received

    def _is_mab_restart(self):  # pylint: disable=missing-docstring
        return self.mab_restart

    def _is_aaa_received(self):  # pylint: disable=missing-docstring
        return self.aaa_received

    def _is_aaa_fail(self):  # pylint: disable=missing-docstring
        return self.aaa_fail

    def _is_aaa_success(self):  # pylint: disable=missing-docstring
        return self.aaa_success

    #
    # State Functionality
    #
    @log_method
    def mab_disabled_state(self):  # pylint: disable=missing-docstring
        self.mab_restart = False

    @log_method
    def eth_received_state(self):  # pylint: disable=missing-docstring
        self.process_ethernet_frame()

    @log_method
    def aaa_request_state(self):  # pylint: disable=missing-docstring
        self.send_aaa_request()

    @log_method
    def aaa_idle_state(self):  # pylint: disable=missing-docstring
        pass

    @log_method
    def aaa_received_state(self):  # pylint: disable=missing-docstring
        self.process_radius_message()

    @log_method
    def aaa_success_state(self):  # pylint: disable=missing-docstring
        self.logger.info(
            'Authentication Passed: MAC is approved for MAB %s', self.src_mac)
        self.handle_success()

    @log_method
    def aaa_failure_state(self):  # pylint: disable=missing-docstring
        self.logger.info(
            'Authentication Failed: MAC is not approved for MAB %s', self.src_mac)
        self.handle_failure()

    # pylint: disable=too-many-arguments

    def __init__(self, radius_output_queue, src_mac, timer_scheduler,
                 auth_handler, failure_handler, log_prefix):
        """

        Args:
            auth_handler (callable): callable that takes input of src_mac. Called on MAB-Success
            failure_handler (callable): callable that takes input of src_mac. Called on MAB-Failure.
            radius_output_queue (Queue): where to put Messages to send to AAA server
            src_mac (MacAddress): MAC address this statemachine (sm) belongs to.
            timer_scheduler (Scheduler): where to put timer events. (useful for Retransmits)
            log_prefix (String): the prefix used when outputting logs
        """
        self.radius_output_messages = radius_output_queue
        self.src_mac = src_mac
        self.timer_scheduler = timer_scheduler
        self.auth_handler = auth_handler
        self.failure_handler = failure_handler
        self.aaa_sent_count = 0
        self.set_timer = None
        self.machine = Machine(model=self, states=MacAuthenticationBypassStateMachine.STATES,
                               transitions=MacAuthenticationBypassStateMachine.TRANSITIONS,
                               queued=True,
                               initial=MacAuthenticationBypassStateMachine.DISABLED)

        self.logger = get_logger(log_prefix)

        self.reset_variables()
        self.port_enabled = True
        self.eth_received = True

    #
    # State Functionalty Helper Functions
    #

    def reset_variables(self):
        """Reset all used state-machine variables"""
        self.aaa_received = False
        self.aaa_fail = False
        self.aaa_success = False
        self.aaa_response_data = None
        self.aaa_request_data = None
        self.eth_received = False
        self.eth_message_data = None
        self.aaa_response_attributes = None

    def event(self, event):
        """Processes an event for the state machine"""
        self.logger.info(
            "Received event: %s with starting state: %s", event.__class__, self.state)

        self.reset_variables()

        # Process Decisions
        if isinstance(event, EventMessageReceived):
            self.event_message_received(event)
        else:
            self.logger.error('MAB State Machine error. Incorrect event received. %s',
                              event.__dict__)

        self.handle_event_received()
        self.logger.info('end state: %s', self.state)

    def handle_success(self):
        """Handle a AAA_Success event"""
        self.logger.info("Successful MAB Authentication. Running Auth Handler")
        self.auth_handler(self.src_mac, str(self.port_id_mac), self.session_timeout,
                          self.aaa_response_attributes)

    def handle_failure(self):
        """Handle a AAA_Failure event"""
        self.logger.info("Failed MAB Authentication. Running Failure Handler")
        self.failure_handler(self.src_mac, str(self.port_id_mac))

    def handle_event_received(self):
        """Main state machine loop"""
        last_state = None
        while self.state != last_state:
            last_state = self.state
            self.process()  # pylint: disable=no-member # pytype: disable=attribute-error

    def event_message_received(self, event):
        """Handle a message received event"""
        if event.port_id:
            self.port_id_mac = event.port_id

        if isinstance(event, EventRadiusMessageReceived):
            self.aaa_received = True
            self.aaa_response_data = event.message
            self.aaa_response_attributes = event.attributes

            if isinstance(self.aaa_response_data, RadiusAccessAccept):
                self.aaa_success = True
            elif isinstance(self.aaa_response_data, RadiusAccessReject):
                self.aaa_fail = True

        else:
            self.eth_received = True
            self.eth_message_data = event.message

    def process_ethernet_frame(self):
        """Perform checks on ethernet frames"""

    def process_radius_message(self):
        """Perform checks on Radius Packets before they're passed to the State Machine"""
        if not isinstance(self.aaa_response_data, RadiusAccessAccept) \
                and not isinstance(self.aaa_response_data, RadiusAccessReject):
            raise Exception("Unexpected Packet Type in MAB state Machine: %s" %
                            self.aaa_response_data.__dict__)

    def send_aaa_request(self):
        """Perform sending a AAA Request"""
        port_id = self.port_id_mac
        ethernet_packet = self.eth_message_data
        src_mac = ethernet_packet.src_mac

        # Build the RADIUS Packet and send
        self.radius_output_messages.put_nowait(
            RadiusQueueMessage(src_mac, src_mac, src_mac,
                               self.radius_state_attribute, port_id))

        self.aaa_sent_count += 1



"""Test Chewie's MAB Functionality"""

import time
import unittest
from base_test import BaseTest

CONFIG_FILENAME = "/tmp/wpasupplicant/wired-tmp.conf"
supplicant_config_template = """   # pylint: disable=invalid-name
ctrl_interface=/tmp/wpa_supplicant
ctrl_interface_group=0
ap_scan=0

network={{
 key_mgmt=IEEE8021X
 eap=MD5
 identity="{0}"
 password="{1}"
 eapol_flags=0
}}"""


# TODO : should really not be so dependent on the radius users file
# TODO : should also test for mab
class RadiusAttributeOnAuthentication(BaseTest):
    """Test Chewie's MAB Functionality"""
    test_name = "RadiusAttributeOnAuthentication"

    @staticmethod
    def create_supplicant_config(identity, password):
        """Create config file with creds."""
        with open(CONFIG_FILENAME, "w+") as tmp_wpa_config:
            tmp_wpa_config.write(supplicant_config_template.format(identity, password))

    def setUp(self):
        """Start Radius and Chewie Servers"""
        super().setUp()
        self.start_radius()
        self.start_chewie()

    def test_smoke_vlan_id(self):
        """Perform MAB using dhclient"""
        self.create_supplicant_config("vlan_id", "microphone")
        self.start_wpa_supplicant('tmp')
        time.sleep(5)
        requirements = ['kwargs : vlan_name : VLAN_100']

        self.check_output(chewie_requirements=requirements)

    def test_smoke_filter_id(self):
        """Perform MAB using dhclient"""
        self.create_supplicant_config("filter_id", "microphone")
        self.start_wpa_supplicant('tmp')
        time.sleep(5)
        requirements = ['kwargs : filter_id : ACL_1']
        self.check_output(chewie_requirements=requirements)

    @unittest.skip("Skipping as not implemented yet.")
    def test_smoke_nas_filter_rule(self):
        """Perform MAB using dhclient"""
        self.create_supplicant_config("filter_rule", "microphone")
        self.start_wpa_supplicant('tmp')
        time.sleep(5)
        requirements = ['kwargs : NAS-Filter-Rule : deny in tcp from any to any']
        self.check_output(chewie_requirements=requirements)



import logging
import sys
import argparse

from chewie.chewie import Chewie


def get_logger(name, log_level=logging.DEBUG):
    """Create and return a logger object"""

    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(log_level)
        logger_handler = logging.StreamHandler(sys.stdout)
        logger_handler.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        logger_handler.setFormatter(formatter)
        logger.addHandler(logger_handler)
    return logger


def auth_handler(address, group_address, *args, **kwargs):
    """Handle successfully authenticated clients"""

    logger = get_logger("CHEWIE")
    logger.info("Authentication successful for address %s on port %s",
                str(address), str(group_address))
    logger.info("Arguments passed from Chewie to Faucet: \n*args: %s\n**kwargs: %s",
                str(args), str(kwargs))


def failure_handler(address, group_address):
    """Handle client failures"""

    logger = get_logger("CHEWIE")
    logger.info("Authentication failed for address %s on port %s",
                str(address), str(group_address))


def logoff_handler(address, group_address):
    """Handle client logoffs"""

    logger = get_logger("CHEWIE")
    logger.info("Logoff successful for address %s on port %s",
                str(address), str(group_address))


def main():
    """Chewie main function, configure and start a chewie process"""

    parser = argparse.ArgumentParser(
        description='Run Chewie 802.1x Authenticator independently of '
                    'Faucet SDN Controller')

    parser.add_argument(
        '-i',
        '--interface',
        dest='interface',
        help='Set the interface for Chewie to listen on - Default: eth0',
        default="eth0")
    parser.add_argument(
        '-ri',
        '--radius_ip',
        dest='radius_ip',
        help='Set the IP Address for the RADIUS Server that Chewie will forward requests to '
             '- DEFAULT: 127.0.0.1',
        default='127.0.0.1')
    parser.add_argument(
        '-rs',
        '--radius_secret',
        dest='radius_secret',
        help='Set the Secret used for connecting to the RADIUS Server - Default: SECRET',
        default='SECRET')
    args = parser.parse_args()

    logger = get_logger("CHEWIE")
    logger.info('Starting Chewie...')

    chewie = Chewie(args.interface, logger, auth_handler, failure_handler, logoff_handler,
                    radius_server_ip=args.radius_ip, radius_server_secret=args.radius_secret)
    chewie.run()


if __name__ == '__main__':
    main()



# pylint: disable=missing-docstring

import unittest

from chewie.auth_8021x import Auth8021x


class Auth8021xTestCase(unittest.TestCase):
    def test_auth_8021x_parses(self):
        packed_message = bytes.fromhex("01000005010100050100000000")
        message = Auth8021x.parse(packed_message)
        self.assertEqual(message.version, 1)
        self.assertEqual(message.packet_type, 0)
        self.assertEqual(len(message.data), 5)

    def test_auth_8021x_packs(self):
        expected_packed_message = bytes.fromhex("010000050101000501")
        message = Auth8021x(version=1, packet_type=0, data=bytes.fromhex("0101000501"))
        packed_message = message.pack()
        self.assertEqual(expected_packed_message, packed_message)



"""Utility Functions"""
import logging
import random
from collections import namedtuple  # pytype: disable=pyi-error


def get_logger(logname):
    """Create and return a logger object."""
    logger = logging.getLogger(logname)
    return logger


def log_method(method):
    """Generate method for logging"""

    def wrapped(self, *args, **kwargs):
        """Method that gets called for logging"""
        self.logger.info('Entering %s' % method.__name__)
        return method(self, *args, **kwargs)

    return wrapped


def get_random_id():  # pylint: disable=missing-docstring
    """pick a random 8 bit EAPOL ID - which cannot be 0."""
    return random.randint(1, 2**8 - 1)


class MessageParseError(Exception):
    """Error for when parsing cannot be successfully completed."""
    pass


class EapQueueMessage(namedtuple('EapQueueMessage',
                                 'message src_mac port_mac')):
    pass


class RadiusQueueMessage(namedtuple('RadiusQueueMessage',
                                    'message src_mac identity state port_mac')):
    pass



#!/usr/bin/env python3

"""Chewie setup script"""

import sys

from setuptools import setup

if sys.version_info < (3,):
    print("""You are trying to install chewie on python {py}

chewie is not compatible with python 2, please upgrade to python 3.6 or newer."""
          .format(py='.'.join([str(v) for v in sys.version_info[:3]])), file=sys.stderr)
    sys.exit(1)

setup(
    name='chewie',
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    python_requires='>=3.6',
    pbr=True
)



"""Module for parsing & packing EAP"""

# Ignore arguments differ on overridden functions due to changing inheritance hierarchies
# pylint: disable=arguments-differ

import struct
from chewie.utils import MessageParseError

EAP_HEADER_LENGTH = 1 + 1 + 2
EAP_TYPE_LENGTH = 1

PARSERS = {}
PARSERS_TYPES = {}


class Eap:
    """Class for parsing & packing EAP messages"""
    REQUEST = 1
    RESPONSE = 2
    SUCCESS = 3
    FAILURE = 4

    IDENTITY = 1
    LEGACY_NAK = 3
    MD5_CHALLENGE = 4
    TLS = 13
    TTLS = 21
    PEAP = 25

    code = None
    packet_id = None
    PACKET_TYPE = None

    @staticmethod
    def parse(packed_message):
        """

        Args:
            packed_message:
        Returns:
            Eap*** object.
        Raises:
            MessageParseError if packed_message cannot be parsed.
        """
        try:
            code, packet_id, length = struct.unpack("!BBH", packed_message[:EAP_HEADER_LENGTH])
        except struct.error as exception:
            raise MessageParseError("unable to unpack EAP header (4 bytes)") from exception

        if code in (Eap.REQUEST, Eap.RESPONSE):
            try:
                packet_type, = struct.unpack("!B",
                                             packed_message[EAP_HEADER_LENGTH:
                                                            EAP_HEADER_LENGTH + EAP_TYPE_LENGTH])
            except struct.error as exception:
                raise MessageParseError("EAP unable to unpack packet_type byte") \
                    from exception

            data = packed_message[EAP_HEADER_LENGTH + EAP_TYPE_LENGTH:length]
            try:
                return PARSERS[packet_type](code, packet_id, data)
            except KeyError as exception:
                raise MessageParseError("EAP packet_type: %s not supported" % packet_type) \
                    from exception
        elif code == Eap.SUCCESS:
            return EapSuccess(packet_id)
        elif code == Eap.FAILURE:
            return EapFailure(packet_id)
        raise MessageParseError("Got Eap packet with bad code: %s" % packed_message)

    def pack(self, packed_body):
        """Pack an EAP Message"""
        header = struct.pack("!BBHB", self.code, self.packet_id,
                             EAP_HEADER_LENGTH + EAP_TYPE_LENGTH + len(packed_body),
                             self.PACKET_TYPE)
        return header + packed_body


def register_parser(cls):
    """Register a packet type to the parser"""
    PARSERS[cls.PACKET_TYPE] = cls.parse
    PARSERS_TYPES[cls.PACKET_TYPE] = cls
    return cls

# pylint: disable=missing-docstring
@register_parser
class EapIdentity(Eap):
    PACKET_TYPE = Eap.IDENTITY

    def __init__(self, code, packet_id, identity):
        self.code = code
        self.packet_id = packet_id
        self.identity = identity

    @classmethod
    def parse(cls, code, packet_id, packed_message):
        """
        Returns:
            EapIdentity.
        Raises:
            MessageParseError if packed message cannot be decoded.
        """
        try:
            identity = packed_message.decode()
        except UnicodeDecodeError as exception:
            raise MessageParseError("%s unable to decode identity" % cls.__name__) from exception
        return cls(code, packet_id, identity)

    def pack(self):
        packed_identity = self.identity.encode()
        return super().pack(packed_identity)

    def __repr__(self):
        return "%s(identity=%s)" % \
               (self.__class__.__name__, self.identity)


@register_parser
class EapMd5Challenge(Eap):
    """EAP MD5-Challenge Packet"""

    PACKET_TYPE = Eap.MD5_CHALLENGE

    def __init__(self, code, packet_id, challenge, extra_data):
        self.code = code
        self.packet_id = packet_id
        self.challenge = challenge
        self.extra_data = extra_data

    @classmethod
    def parse(cls, code, packet_id, packed_message):
        """
        Returns:
            EapMd5Challenge.
        Raises:
            MessageParseError if cannot unpack packed_message
        """
        try:
            value_length, = struct.unpack("!B", packed_message[:1])
        except struct.error as exception:
            raise MessageParseError("%s unable to unpack first byte" % cls.__name__) \
                from exception
        challenge = packed_message[1:1 + value_length]
        extra_data = packed_message[1 + value_length:]
        return cls(code, packet_id, challenge, extra_data)

    def pack(self):
        value_length = struct.pack("!B", len(self.challenge))
        packed_md5_challenge = value_length + self.challenge + self.extra_data
        return super().pack(packed_md5_challenge)

    def __repr__(self):
        return "%s(challenge=%s, extra_data=%s)" % \
               (self.__class__.__name__, self.challenge, self.extra_data)


class EapSuccess(Eap):
    """EAP Success Packet"""
    def __init__(self, packet_id):
        self.packet_id = packet_id

    @classmethod
    def parse(cls, packet_id):
        return cls(packet_id)

    def pack(self):
        return struct.pack("!BBH", Eap.SUCCESS, self.packet_id, EAP_HEADER_LENGTH)

    def __repr__(self):
        return "%s(packet_id=%s)" % \
               (self.__class__.__name__, self.packet_id)


class EapFailure(Eap):
    """EAP Failure Packet"""
    def __init__(self, packet_id):
        self.packet_id = packet_id

    @classmethod
    def parse(cls, packet_id):
        return cls(packet_id)

    def pack(self):
        return struct.pack("!BBH", Eap.FAILURE, self.packet_id, EAP_HEADER_LENGTH)

    def __repr__(self):
        return "%s(packet_id=%s)" % \
               (self.__class__.__name__, self.packet_id)


@register_parser
class EapLegacyNak(Eap):
    """EAP Legacy-NAK Packet"""
    PACKET_TYPE = Eap.LEGACY_NAK

    def __init__(self, code, packet_id, desired_auth_types):
        self.code = code
        self.packet_id = packet_id
        self.desired_auth_types = desired_auth_types

    @classmethod
    def parse(cls, code, packet_id, packed_msg):
        """
        Returns:
            EapLegacyNak.
        Raises:
            MessageParseError if cannot unpack packed_message
        """
        value_len = len(packed_msg)
        try:
            desired_auth_types = struct.unpack("!%ds" % value_len, packed_msg)
        except struct.error as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception
        return cls(code, packet_id, desired_auth_types)

    def pack(self):
        packed_legacy_nak = struct.pack("!%ds" % len(self.desired_auth_types),
                                        *self.desired_auth_types)  # pytype: disable=wrong-arg-types
        return super().pack(packed_legacy_nak)

    def __repr__(self):
        return "%s(packet_id=%s, desired_auth_types=%s)" % \
               (self.__class__.__name__, self.packet_id, self.desired_auth_types)


class EapTLSBase(Eap):
    """EAPTLS & EAPTTLS have the same packet format."""

    def __init__(self, code, packet_id, flags, extra_data):
        self.code = code
        self.packet_id = packet_id
        self.flags = flags
        self.extra_data = extra_data

    @classmethod
    def parse(cls, code, packet_id, packed_msg):
        """
        Returns:
            A child of EapTLSBase e.g. EapTLS, EAPTTLS.
        Raises:
            MessageParseError if cannot unpack packed_message
        """
        value_len = len(packed_msg)
        fmt_str = "!B"
        if value_len > 1:
            fmt_str += "%ds" % (value_len - 1)
        try:
            unpacked = struct.unpack(fmt_str, packed_msg)
        except struct.error as exception:
            raise MessageParseError("%s unable to unpack" % cls.__name__) from exception
        extra_data = b""
        if value_len > 1:
            flags, extra_data = unpacked
        else:
            flags = unpacked[0]

        return cls(code, packet_id, flags, extra_data)

    def pack(self):
        if self.extra_data:
            packed = struct.pack("!B%ds" % len(self.extra_data), self.flags, self.extra_data)
        else:
            packed = struct.pack("!B", self.flags)
        return super().pack(packed)

    def __repr__(self):
        return "%s(packet_id=%s, flags=%s, extra_data=%s)" % \
               (self.__class__.__name__, self.packet_id, self.flags, self.extra_data)


@register_parser
class EapTLS(EapTLSBase):
    """EAP TLS Packet"""
    PACKET_TYPE = Eap.TLS


@register_parser
class EapTTLS(EapTLSBase):
    """EAP TTLS Packet"""
    PACKET_TYPE = Eap.TTLS


@register_parser
class EapPEAP(EapTLSBase):
    """PEAP Packet"""
    PACKET_TYPE = Eap.PEAP



"""This module is used to allow parsing and packing of Ethernet Packets"""
import struct

from chewie.mac_address import MacAddress
from chewie.utils import MessageParseError

ETHERNET_HEADER_LENGTH = 6 + 6 + 2


class EthernetPacket:
    """Packet/parsers for an IEEE 802.3 Ethernet frame"""

    def __init__(self, dst_mac, src_mac, ethertype, data):
        self.dst_mac = dst_mac
        self.src_mac = src_mac
        self.ethertype = ethertype
        self.data = data

    @classmethod
    def parse(cls, packed_message):
        """construct an EthernetPacket from a packed_message
        Args:
            packed_message (bytes):
        Returns:
            EthernetPacket
        Raises:
            MessageParseError: if packed_message cannot be successfully parsed.
        """
        try:
            dst_mac, src_mac, ethertype = struct.unpack("!6s6sH",
                                                        packed_message[:ETHERNET_HEADER_LENGTH])
        except struct.error as exception:
            raise MessageParseError("Unable to parse Ethernet header (14bytes)") from exception
        data = packed_message[ETHERNET_HEADER_LENGTH:]
        return cls(MacAddress(dst_mac), MacAddress(src_mac), ethertype, data)

    def pack(self):
        """Pack up the ethernet packet to bytes.
        Returns:
            bytes
        """
        header = struct.pack("!6s6sH", self.dst_mac.address, self.src_mac.address, self.ethertype)
        return header + self.data

    def __repr__(self):
        return "%s(dst_mac=%s, src_mac=%s, ethertype=0x%04X)" % \
               (self.__class__.__name__, self.dst_mac, self.src_mac, self.ethertype)



"""Test Chewie's MAB Functionality"""

import time
from base_test import BaseTest


class MabTest(BaseTest):
    """Test Chewie's MAB Functionality"""
    test_name = "MabTest"

    def setUp(self):
        """Start Radius and Chewie Servers"""
        super().setUp()
        self.start_radius()
        self.start_chewie()

    def test_smoke_mab(self):
        """Perform MAB using dhclient"""
        self.start_dhclient()
        last_err = None
        for _ in range(3):
            time.sleep(5)
            try:
                self.check_output()
                return
            except AssertionError as err:
                last_err = err
                continue
        self.fail(last_err)



"""This Module provides the Abstract Design Requirements for a State Machine in Chewie"""
from transitions.extensions import GraphMachine


class AbstractStateMachine:
    """This Class provides the Abstract Design Requirements for a State Machine in Chewie"""

    PROGRESS_STATES = []

    SUCCESS_STATES = []
    FAILURE_STATES = []
    COMPLETION_STATES = FAILURE_STATES + SUCCESS_STATES
    INITIAL_STATE = None
    STATES = COMPLETION_STATES + PROGRESS_STATES

    ERROR_TRANSTIONS = []

    CORE_TRANSITIONS = []
    TRANSITIONS = CORE_TRANSITIONS + ERROR_TRANSTIONS
    port_enabled = None
    state = None

    def is_in_progress(self):
        """
        Returns true if the state machine is currently in progress
        """
        return self.port_enabled and self.state not in self.COMPLETION_STATES

    def is_success(self):
        """
        Returns true if the state machine currently in a successful completion state and enabled
        """
        return self.port_enabled and self.state in self.SUCCESS_STATES

    @classmethod
    def build_state_graph(cls, filename):
        "Build a graphc representation of the state machine and store in 'filename'.png"
        model = type('model', (object,), {})()
        GraphMachine(model=model, states=cls.STATES,
                     title=cls.__name__,
                     transitions=cls.CORE_TRANSITIONS,
                     queued=True,
                     initial=cls.INITIAL_STATE)
        # pylint: disable=no-member
        # pytype: disable=attribute-error
        model.get_graph().draw(filename, prog='dot')



"""This module is used as a base for other integration tests"""

# To extract the logs from the Docker instance, override /tmp/logs


import inspect
import os
import shutil
import signal
import subprocess
import unittest
import tempfile
from collections import namedtuple
import logging
import sys

from chewie.chewie import Chewie


def get_logger(name, file=sys.stdout, log_level=logging.DEBUG):
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(log_level)
        handler = logging.StreamHandler(file)
        handler.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


VethLink = namedtuple('VethLink', 'name ip mac')

CHEWIE_SUPPLICANT = VethLink('chewie_supp', '192.168.20.1', '8e:00:00:00:01:01')
# TODO - Add namespaces
CHEWIE_RADIUS = VethLink('chewie_radius', '192.168.21.1', '8e:00:00:00:02:01')
SUPPLICANT = VethLink('supplicant', '192.168.20.2', '8e:00:00:00:01:02')
RADIUS = VethLink('radius', '127.0.0.1', '8e:00:00:00:02:02')

# NOTE: DO NOT CHANGE THESE VALUES UNLESS CHANGES ARE MADE TO THE FREERADIUS CONFIGURATION FILES
RADIUS_IP = RADIUS.ip
RADIUS_PORT = "1812"
RADIUS_SECRET = "SECRET"

IP_LINK_PAIRS = {
    CHEWIE_SUPPLICANT: SUPPLICANT,
    CHEWIE_RADIUS: RADIUS,
}


LOG_DIR = '/tmp/logs/'
os.makedirs(LOG_DIR, exist_ok=True)

HANDLER_COUNTS = {}

CHEWIE_ROOT = os.environ.get('CHEWIE_ROOT', None)

if not CHEWIE_ROOT:
    CHEWIE_ROOT = '/chewie-src/'

CHEWIE_CONF_DIR = CHEWIE_ROOT + '/etc/'

def auth_handler(address, group_address, *args, **kwargs):  # pylint: disable=missing-docstring
    logger = logging.getLogger('CHEWIE')
    logger.info("Authentication successful for address %s on port %s",
                str(address), str(group_address))
    logger.info("Arguments passed from Chewie to Faucet: \n*args: %s", str(args))

    if kwargs:
        for key, value in kwargs.items():
            logger.info("kwargs : %s : %s", str(key), str(value))

    HANDLER_COUNTS['auth_handler'] += 1  # pytype: disable=key-error


def failure_handler(address, group_address):  # pylint: disable=missing-docstring
    logger = logging.getLogger('CHEWIE')
    logger.info("Authentication failed for address %s on port %s",
                str(address), str(group_address))

    HANDLER_COUNTS['failure_handler'] += 1  # pytype: disable=key-error


def logoff_handler(address, group_address):  # pylint: disable=missing-docstring
    logger = logging.getLogger('CHEWIE')
    logger.info("Logoff successful for address %s on port %s",
                str(address), str(group_address))

    HANDLER_COUNTS['logoff_handler'] += 1  # pytype: disable=key-error


class BaseTest(unittest.TestCase):
    """
    This class can be used to hold common functionality of integration tests for Chewie
    Inherit from this class to have an environment set up for the tests to run in.
    """
    active_processes = []
    chewie_pid = None
    freeradius_log = None
    wpa_supplicant_log = None
    current_log_dir = None

    test_name = "BaseTest"

    @classmethod
    def setUpClass(cls):
        cls.prepare_freeradius()
        cls.prepare_wpa_supplicant()

    def setUp(self):
        """Setup environment for tests to start processes"""

        self.active_processes = []
        self.freeradius_log = None
        self.wpa_supplicant_log = None

        HANDLER_COUNTS = {
            'auth_handler': 0,
            'logoff_handler': 0,
            'failure_handler': 0
        }

        for link1, link2 in IP_LINK_PAIRS.items():
            self.run_command_and_wait(
                "ip link add {} type veth peer name {}".format(link1.name, link2.name))
            for link in [link1, link2]:
                self.run_command_and_wait(
                    "ip link set dev {} address {}".format(link.name, link.mac))
                self.run_command_and_wait("ip link set {} up".format(link.name))

        self.open_logs()

    def tearDown(self):
        """Close Logs and Kill Opened Processes"""
        self.close_logs()

        if self.chewie_pid != 0:
            os.kill(self.chewie_pid, signal.SIGKILL)

        for proc in self.active_processes:
            os.kill(proc.pid, signal.SIGKILL)
            proc.wait()

        for link1, _ in IP_LINK_PAIRS.items():
            self.run_command_and_wait("ip link del {}".format(link1.name))

    def open_logs(self):
        """Open Logs for Processes"""
        self.current_log_dir = tempfile.mkdtemp(prefix='chewie-' + self.test_name + '-',
                                                dir='/tmp/logs') + "/"
        print('Logging test results in {}'.format(self.current_log_dir))
        print(os.path.join(self.current_log_dir + "wpa_supplicant.log"))
        self.freeradius_log = open(os.path.join(self.current_log_dir, "freeradius.log"), "w+")
        self.wpa_supplicant_log = open(os.path.join(self.current_log_dir + "wpa_supplicant.log"),
                                       "w+")

    def close_logs(self):
        """Close Process Logs"""
        self.freeradius_log.close()
        self.wpa_supplicant_log.close()

    def run_command_and_wait(self, command, output_file=None):  # pylint: disable=no-self-use
        """Run a command and wait for the process to complete"""
        if output_file:
            child = subprocess.Popen(command.split(), stdout=output_file)
        else:
            child = subprocess.Popen(command.split())

        result = child.wait()
        if result != 0:
            raise Exception(
                "Command returned with a non-zero exit code. Code: {}, Command: {}".format(
                    str(result), command))

    def run_command_and_detach(self, command, output_file=None):
        """Run a command and return the process"""
        if output_file:
            child = subprocess.Popen(command.split(), stdout=output_file)
        else:
            child = subprocess.Popen(command.split())
        self.active_processes.append(child)
        return child

    def start_radius(self):
        """Start Radius Server"""
        # NOTE: RADIUS PORT and IP have not been set due to it
        # skipping sections of the radiusd.conf file when given.
        return self.run_command_and_detach("freeradius -X -l stdout", self.freeradius_log)

    def start_chewie(self):
        """Start Chewie Server"""

        self.chewie_pid = os.fork()
        if self.chewie_pid == 0:
            file = open(os.path.join(self.current_log_dir + "chewie.log"), 'w+')
            logger = get_logger('CHEWIE', file)
            logger.info('Starting chewie.')
            chewie = Chewie(CHEWIE_SUPPLICANT.name, logger, auth_handler,
                            failure_handler, logoff_handler,
                            radius_server_ip=RADIUS_IP, radius_server_secret=RADIUS_SECRET)
            chewie.run()

    def start_wpa_supplicant(self, eap_method):
        """Start WPA_Supplicant / EAP Client"""
        return self.run_command_and_detach(
            "wpa_supplicant -dd -c/tmp/wpasupplicant/wired-{}.conf -i{} -Dwired".format(
                eap_method, SUPPLICANT.name), self.wpa_supplicant_log)

    def start_dhclient(self):
        """Start dhclient on the MAB port"""
        return self.run_command_and_detach("dhclient -i {}".format(SUPPLICANT.name))

    def check_output(self, **kwargs):  # pylint: disable=no-self-use
        """Check the output of the Log Files to verify state of system"""

        with open(os.path.join(self.current_log_dir + "chewie.log"), "r") as file:
            chewie_log = file.read()

        chewie_requirements = kwargs.get("chewie_requirements", None)
        if chewie_requirements:
            for requirement in chewie_requirements:
                assert requirement in chewie_log, "Unable to find {} in chewie logs".format(
                    requirement, )

        currentframe = inspect.currentframe()
        assert currentframe

        if currentframe:
            assert "Authentication successful" in chewie_log, "Authentication failed for {}".format(
                currentframe.f_back.f_code.co_name)


    @staticmethod
    def prepare_freeradius():
        chewie_rad_dir = CHEWIE_CONF_DIR + "freeradius/"
        if os.path.isfile('/etc/freeradius/users'):
            # Assume we are dealing with freeradius < 3
            radius_config_base = '/etc/freeradius/'
        else:
            # Assume we are dealing with freeradius >=3 configuration
            freerad_version = os.popen(
                r'freeradius -v | egrep -o -m 1 "Version ([0-9]\.[0.9])"').read().rstrip()
            freerad_major_version = freerad_version.split(' ')[1]
            radius_config_base = '/etc/freeradius/%s/' % freerad_major_version

        try:
            # Copy files
            file_map = {
                chewie_rad_dir + 'clients.conf': radius_config_base,
                chewie_rad_dir + 'users': radius_config_base,
                chewie_rad_dir + 'default/eap': radius_config_base + 'mods-available/',
                chewie_rad_dir + 'default/inner-eap': radius_config_base + 'mods-available/',
                chewie_rad_dir + 'default/tls': radius_config_base + 'sites-available/',
            }

            for src, dst in file_map.items():
                shutil.copy(src, dst)

            # Copy Folder
            folder_map = {
                chewie_rad_dir + 'certs': radius_config_base + 'certs',
            }
            for src, dst in folder_map.items():
                if os.path.exists(dst) and os.path.isdir(dst):
                    shutil.rmtree(dst)
                shutil.copytree(src, dst)

        except OSError as err:
            print("Unable to copy FreeRadius files into place.", file=sys.stderr)
            raise err

    @staticmethod
    def prepare_wpa_supplicant():
        folder_map = {
            CHEWIE_CONF_DIR + 'wpasupplicant/': '/tmp/wpasupplicant',
            CHEWIE_CONF_DIR + 'wpasupplicant/cert': '/tmp/cert'
        }
        for src, dst in folder_map.items():
            if os.path.exists(dst) and os.path.isdir(dst):
                shutil.rmtree(dst)
            shutil.copytree(src, dst)



"""This module provides a way to parse and pack and EAPOL Packet using Auth8021x"""

import struct
from chewie.utils import MessageParseError


def hwaddr_to_string(hwaddr):  # pylint: disable=missing-docstring
    return ":".join(["%02x" % x for x in hwaddr])


AUTH_8021X_HEADER_LENGTH = 1 + 1 + 2


class Auth8021x:
    """basically a EAPOL packet"""

    def __init__(self, version, packet_type, data):
        self.version = version
        self.packet_type = packet_type
        self.data = data

    @classmethod
    def parse(cls, packed_message):
        """construct an Auth8021x from a packed_message
        Args:
            packed_message (bytes):
        Returns:
            Auth8021x
        Raises:
            MessageParseException: if packed_message cannot be parsed successfully.
        """
        try:
            version, packet_type, length = struct.unpack("!BBH",
                                                         packed_message[:AUTH_8021X_HEADER_LENGTH])
        except struct.error as exception:
            raise MessageParseError("Auth8021x unable to parse first 4 bytes") from exception
        data = packed_message[AUTH_8021X_HEADER_LENGTH:AUTH_8021X_HEADER_LENGTH + length]
        return cls(version, packet_type, data)

    def pack(self):
        """Pack up the EAPOL packet to bytes.
        Returns:
            bytes
        """
        header = struct.pack("!BBH", self.version, self.packet_type, len(self.data))
        return header + self.data

    def __repr__(self):
        return "%s(version=%s, packet_type=%s, data=%s)" % \
               (self.__class__.__name__, self.version, self.packet_type, self.data)

    def __str__(self):
        return "%s<packet_type=%d, data=%s>" % \
               (self.__class__.__name__, self.packet_type, self.data)
